import time

from eras.config.config import config
from eras.models.simple_stream_inference_callable import (SimpleStreamInferenceCallable, LLMTextReceivedCallable,
                                                          LLMResponseCompletedCallable)


class SimpleLLM:
    def __init__(self, openai_client):

        self.client = openai_client

    def stream_inference(self, prompt: str, handle_on_text_received: LLMTextReceivedCallable,
                         handle_response_complete: LLMResponseCompletedCallable) -> None:
        # start_time_seconds = time.time()
        system_message = f"""
        You are friendly AI assistant that runs in a terminal for the operating system {config.get_user_operating_system()}.  
        All responses you provide will be shown in a terminal.  You are an expert in ensuring your response text is plain text, and does not use markdown.
        """
        completion = self.client.chat.completions.create(
            model=config.get_model_name(),
            messages=[
                {"role": "system", "content": system_message},
                {"role": "user", "content": prompt}
            ],
            stream=True,  # Whether to stream the response or return the full response at once
        )

        for chunk in completion:
            text = chunk.choices[0].delta.content
            if text is not None:
                handle_on_text_received(text)

        handle_response_complete()
        # print()
        # print(f"answer received in {time.time() - start_time_seconds} seconds.")

    def get_simple_stream_inference_callable(self) -> SimpleStreamInferenceCallable:
        return lambda prompt, handle_on_text_received, handle_response_complete: (
            self.stream_inference(prompt, handle_on_text_received, handle_response_complete))



import json
from eras.decorators.chatgpt_tool_data import chatgpt_tool_data
from eras.services.llm_callable_functions.callable_function_service_base import CallableFunctionServiceBase


# Demo service which provides ChatGPT the ability to call method get_user_details, which supplies a hard-coded result.
class TerminalCommand(CallableFunctionServiceBase):
    def __init__(self):
        super().__init__()

    @chatgpt_tool_data({
        "type": "function",
        "function": {
            "name": "execute_terminal_command",
            "description": "Executes a terminal command",
            "parameters": {
                "type": "object",
                "properties": {
                    "command_to_run": {
                        "type": "string",
                        "description": "A string of the command to run in the terminal"
                    },
                },
                "required": ["command_to_run"]
            }
        },
    })
    def execute_terminal_command(self, args):
        arguments = json.loads(args)
        print(
            f"execute_terminal_command function called with arguments first_name: {arguments['command_to_run']}")





# decorator so we can use attributes to describe the function behavior in a way in which chatgpt understands.
# Example:
#     @get_tool_data({
#         "type": "function",
#         "function": {
#             "name": "get_user_details",
#             "description": "Retrieves information for a user based on the combination of their first and last names.  It returns information about the user's age, location, and profession.",
#             "parameters": {
#                 "type": "object",
#                 "properties": {
#                     "first_name": {
#                         "type": "string",
#                         "description": "First name of the user to get details of."
#                     },
#                     "last_name": {
#                         "type": "string",
#                         "description": "Last name of the user to get details of."
#                     }
#                 },
#                 "required": ["first_name", "last_name"]
#             }
#         },
#     })
#     def get_user_details(self, args):
#         ...
def chatgpt_tool_data(tool_data_dict):
    def decorator(func):
        # Attach the tool data to the function
        func.tool_data = tool_data_dict
        return func

    return decorator


class Message:
    def __init__(self, message_text: str, role: str = "user", tool_call_id: str = None, function_name: str = None,
                 raw_chatgpt_chat_completion_response=None, tool_calls=None):
        self.function_name = function_name
        self.tool_call_id = tool_call_id
        self.role = role
        self.message_text = message_text
        self.raw_chatgpt_chat_completion_response = raw_chatgpt_chat_completion_response
        # "tool_calls": [
        #     {
        #         "id": "call_2bsu34b16on3bRm1QFfzhZFq",
        #         "type": "function",
        #         "function": {
        #             "name": "get_user_details",
        #             "arguments": "{\n  \"first_name\": \"Jason\",\n  \"last_name\": \"McAffee\"\n}"
        #         }
        #     }
        # ]
        self.tool_calls = tool_calls  # not parsing this out for right now

    # def to_chatgpt_format(self):
    #     return {
    #         "role": self.role,
    #         "content": self.message_text,
    #         "tool_call_id": self.tool_call_id,
    #         "name": self.function_name,
    #     }

    def to_chatgpt_format(self):
        return {
            key: value for key, value in [
                ("role", self.role),
                ("content", self.message_text),
                ("tool_call_id", self.tool_call_id),
                ("name", self.function_name),
                ("tool_calls", self.tool_calls),
            ] if value is not None
        }



from typing import Callable

# when we receive text from an LLM
LLMTextReceivedCallable = Callable[[str], None]
# when we receive notification that an LLM response has completed
LLMResponseCompletedCallable = Callable[[], None]
# AI bot communicates to an LLM through this abstraction
SimpleStreamInferenceCallable = Callable[[str, LLMTextReceivedCallable, LLMResponseCompletedCallable], None]



import requests
import uuid
import threading
from eras.config.config import config
from ga4mp import GtagMP


def send_analytics_event(event_name, event_params=''):
    def send_event():
        try:
            measurement_id = config.get_google_analytics_measurement_id()
            api_secret = config.get_google_analytics_secret()
            client_id = str(uuid.uuid4())

            message = f"id: {config.get_system_identifier()}, operating_system: {config.get_user_operating_system()}, data: {event_params}"
            # print(message)

            # Initialize GtagMP client
            ga_client = GtagMP(api_secret=api_secret, measurement_id=measurement_id, client_id=client_id)

            # Prepare event data
            event_data = {
                'name': event_name,
                'params': {
                    'message': message
                }
            }

            # Send event
            ga_client.send([event_data])

            # print(f'Event sent: {event_data}')
        except Exception as e:
            pass
            # print(f'Exception sending event: {e}')

    # Run the send_event function in a separate thread
    thread = threading.Thread(target=send_event)
    thread.start()

# def send_analytics_event(event_name, event_params=''):
#     def send_event():
#         try:
#             measurement_id = config.get_google_analytics_measurement_id()
#             api_secret = config.get_google_analytics_secret()
#
#             client_id = str(uuid.uuid4())
#
#             message = f'''id: {config.get_system_identifier()}, operating_system: {config.get_user_operating_system()}, data: {event_params}'''
#             print(message)
#             payload = {
#                 'client_id': client_id,
#                 'events': [
#                     {
#                         'name': event_name,
#                         'params': message
#                     }
#                 ]
#             }
#             response = requests.post(
#                 f'https://www.google-analytics.com/mp/collect?measurement_id={measurement_id}&api_secret={api_secret}',
#                 json=payload
#             )
#             # print('analytics event sent')
#             print(response)
#         except Exception as e:
#             print('exception sending event: ')
#             print(e)
#             pass  # Swallow all exceptions
#
#     # Run the send_event function in a separate thread
#     thread = threading.Thread(target=send_event)
#     thread.start()




import time
import uuid

from eras.config.config import config
from eras.models.conversation import Conversation

# system_prompt = "You are a helpful assistant that answers questions accurately, without making up facts."
model = "gpt-3.5-turbo"


# Agent which:
# - provides an inference function which accepts a question to send to chatgpt and responds with the answer.
# - maintains a conversation, which is a collection of messages that have been sent to and received from chatgpt.
# - determines if chatgpt has requested that a local function has been called.
# - executes the local function and sends the result to chatgpt (via conversation messages)
class TerminalLlamaAgent:
    def __init__(self, openai_client):

        self.openai_client = openai_client


    def get_system_prompt(self):
        user_os = config.get_user_operating_system()
        system_prompt = '''
You are an expert in providing a Natural Language Interface that converts questions and/or instructions into terminal commands.        
You are renowned for your ability to take a user's question or instruction and create a valid terminal command that works every time.
You do not use preamble, explanation, or pleasantries.  You only respond with valid terminal commands that work with the user's operating system.
If there is no way to create a valid terminal command, simply say 'No shell command can facilitate your request'.
Do not surround the terminal command with any string indicators, like `, ', or ".
Here are some example question/command and what your response should look like:

Mac OS Example 1:
Question/Command: Show me the contents of my Documents folder.
Response: ls ~/Documents

Mac OS Example 2:
Question/Command: Create a new directory called 'Projects' in my home folder.
Response: mkdir ~/Projects

Mac OS Example 3:
Question/Command: Delete the file named 'old_report.pdf' from my Desktop.
Response: rm ~/Desktop/old_report.pdf

Mac OS Example 4:
Question/Command: Find all files containing the word 'budget' in my Documents folder.
Response: grep -r "budget" ~/Documents

The user's Operating System is: 
        ''' + user_os
        return system_prompt

    def inference(self, question=None, conversation=None, start_time_seconds=None):
        # print('------- inference ------------')
        # start a new conversation if needed, which tracks all messages to and from chatgpt
        conversation = self.ensure_conversation_exists(conversation)

        # measure how long it takes to complete
        start_time_seconds = start_time_seconds if start_time_seconds is not None else time.time()

        # initial call will have a user question, but some calls will be to pass local function results back to chatgpt.
        conversation.add_user_question_to_messages_if_applicable(question)

        # get all messages in the conversation history to pass to chatgpt
        messages = conversation.get_messages_in_chatgpt_format()
        # print(f"sending messages: {messages}")

        # call chatgpt
        # response = self.openai_client.ChatCompletion.create(model=model, messages=messages, tools=self.tools)
        response = self.openai_client.chat.completions.create(
            model=config.get_model_name(),
            messages=messages,
            # tools=self.tools,
            stream=False, # Whether to stream the response or return the full response at once
            # max_tokens=8000,  # The maximum number of tokens to generate in the completion
            # temperature=0.5,  # The temperature of the model, controlling the randomness of the output
            # top_p=1.0,  # The nucleus sampling parameter, controlling the diversity of the output
            # n=1,  # The number of completions to generate
            # # stop=None,  # A list of stop sequences to stop the completion at
            # stop=None,
            # presence_penalty=0.0,  # The penalty for generating tokens that appear in the prompt
            # frequency_penalty=0.0,  # The penalty for generating tokens that appear frequently in the prompt
            # logit_bias={},  # A dictionary of token IDs and associated biases to apply to the logits
        )
        # print(response)

        # add the response to our conversation.  we may need to pass it back to chatgpt if a function is called.
        conversation.add_llama_response_to_messages(response)

        response_message = conversation.get_last_message()
        # print(response_message.message_text)
        return response_message.message_text

    def ensure_conversation_exists(self, conversation: Conversation):
        if conversation is None:
            conversation = Conversation(conversation_id=str(uuid.uuid4()), system_prompt=self.get_system_prompt())
        return conversation




import os
import sys
import subprocess
import re

def env_var_exists(var_name):
    if os.name == 'nt':  # Windows
        return os.getenv(var_name) is not None
    else:  # Unix-based systems
        home = os.path.expanduser("~")
        shell_config_files = [".bashrc", ".zshrc", ".bash_profile"]

        for config_file in shell_config_files:
            config_path = os.path.join(home, config_file)
            if os.path.exists(config_path):
                with open(config_path, "r") as file:
                    # print(file.read())
                    # print(f'{var_name} is in file? : {var_name in file.read()}')
                    if var_name in file.read():
                        # print(file.read())
                        # print(f'{var_name} is in {file.read()} {config_path}')
                        return True
        return False

def set_env_var(env_var_name, value):
    env_var_command = f"{env_var_name}={value}"
    print(f"setting env var: {env_var_command}")
    os.environ[env_var_name] = value
    set_env_var_to_shell(env_var_command)

def set_env_var_to_shell(env_var_command):
    if os.name == 'nt':  # Windows
        print('setting windows shell env var')
        var_name, var_value = env_var_command.split('=', 1)
        os.system(f'setx {var_name} {var_value}')

    else:  # Unix-based systems
        home = os.path.expanduser("~")
        shell_config_files = [".bashrc", ".zshrc", ".bash_profile"]
        var_name, var_value = env_var_command.split('=', 1)
        export_command = f"export {env_var_command}"

        for config_file in shell_config_files:
            config_path = os.path.join(home, config_file)
            if os.path.exists(config_path):
                with open(config_path, "r") as file:
                    lines = file.readlines()

                with open(config_path, "w") as file:
                    var_found = False
                    for line in lines:
                        if re.match(f"^export {var_name}=", line):
                            file.write(f"{export_command}\n")
                            var_found = True
                        else:
                            file.write(line)
                    if not var_found:
                        file.write(f"\n{export_command}\n")

def ensure():
    if not env_var_exists("ERAS_OPENAI_KEY"):
        openai_key = input("Please enter your OpenAI API key (ERAS_OPENAI_KEY): ").strip()
        if not openai_key:
            print("Error: OpenAI API key cannot be empty.")
            sys.exit(1)
        env_var_command = f"ERAS_OPENAI_KEY={openai_key}"
        os.environ['ERAS_OPENAI_KEY'] = openai_key
        set_env_var_to_shell(env_var_command)
        print("Environment variable 'ERAS_OPENAI_KEY' added to environment.")

        if os.name != 'nt':  # Unix-based systems
            reload_profile(get_profile_file())
        else:
            print('======= PLEASE CLOSE THIS TERMINAL WINDOW AND OPEN A NEW ONE =========')

def get_profile_file():
    if os.name == 'nt':  # Windows
        return None  # No equivalent profile file
    else:  # Unix-based systems
        shell = os.environ.get('SHELL', '')
        if 'zsh' in shell:
            return os.path.expanduser('~/.zshrc')
        elif 'bash' in shell:
            bash_profile = os.path.expanduser('~/.bash_profile')
            if os.path.exists(bash_profile):
                return bash_profile
            else:
                return os.path.expanduser('~/.bashrc')
        else:
            print("Unsupported shell type.  Close this terminal windows and open a new one.")

def reload_profile(profile_file):
    if os.name != 'nt':  # Unix-based systems
        shell = os.environ.get('SHELL', '')
        if 'zsh' in shell:
            os.execvp('zsh', ['zsh', '-c', f'source {profile_file} && exec zsh'])
        elif 'bash' in shell:
            os.execvp('bash', ['bash', '-c', f'source {profile_file} && exec bash'])
        else:
            print(f"Please manually source your profile file: {profile_file}")
    else:
        print("Please restart your Command Prompt or PowerShell session to apply the changes.")

if __name__ == "__main__":
    ensure()




import re

from eras.agents.simple_llm import SimpleLLM
from eras.config.config import config
from eras.services.google_analytics import send_analytics_event
from eras.services.shell_command_service import ShellCommandService
import openai

from eras.services.user_config_service import UserConfigService
import httpx

class UserCommandService:
    def __init__(self):
        self.openai_client = openai.OpenAI(
            base_url=config.get_eras_base_url(),
            api_key=config.get_eras_open_ai_key()
        )
        self.shell_command_service = ShellCommandService(self.openai_client)
        self.simple_llm = SimpleLLM(self.openai_client)
        self.user_config_service = UserConfigService()

    def handle_request(self, user_input: str):
        if user_input.startswith("/help"):
            self.handle_help_request(user_input)
        elif user_input.startswith("/chat"):
            self.handle_chat_request(user_input)
        elif user_input.startswith("/config"):
            self.handle_config_request(user_input)
        else:
            self.handle_shell_nli_request(user_input)

    def handle_help_request(self, user_input):
        send_analytics_event('help')
        self.user_config_service.print_eras_logo()
        self.user_config_service.show_commands()

    def handle_chat_request(self, user_input):
        send_analytics_event('chat')
        match = re.match(r'^/chat\s+(.*)', user_input)
        prompt = match.group(1) if match else None

        def handle_text_received(text):
            print(text, end="")

        def handle_response_complete():
            print()

        self.simple_llm.stream_inference(prompt, handle_text_received, handle_response_complete)


    def handle_shell_nli_request(self, user_input):
        send_analytics_event('nli')
        self.shell_command_service.handle_prompt(user_input)

    def handle_config_request(self, user_input):
        send_analytics_event('config')
        self.user_config_service.prompt_for_all_configs()




# DOESNT WORK - cant get it to return valid json. It's always missing a closing }.
# It does do a good job of calling the correct function though.
# import time
# import uuid
# import json
# from typing import List, Tuple, Set, Dict
#
# from src.factories.function_details_factory import FunctionDetailsFactory
# from src.models.conversation import Conversation
# from src.services.llm_callable_functions.terminal_command import TerminalCommand
# from src.services.llm_callable_functions.user_details import UserDetails
#
# # system_prompt = "You are a helpful assistant that answers questions accurately, without making up facts."
# model = ""
#
#
# # Agent which:
# # - provides an inference function which accepts a question to send to chatgpt and responds with the answer.
# # - maintains a conversation, which is a collection of messages that have been sent to and received from chatgpt.
# # - determines if chatgpt has requested that a local function has been called.
# # - executes the local function and sends the result to chatgpt (via conversation messages)
# class LlamaTerminalCommandFunctionsAgent:
#     def __init__(self, openai_client):
#         print('init')
#         self.callable_function_service = TerminalCommand()
#         self.tools = [] + self.callable_function_service.get_tools()
#         self.openai_client = openai_client
#
#
#     def get_system_prompt(self):
#         tools_as_string = json.dumps(self.tools, indent=4)
#         system_prompt = '''
# You are an expert in taking a user's question or instruction and calling the appropriate function to run a terminal command.
# You are an expert at communicating in only the json format to help determine which function should be called, based on a user's question.
# You are an expert at always providing valid json responses, ensuring that all opening { and [ have corresponding closing } and ].
# You do not use preamble, explanation, or pleasantries.  You only respond with json strings, no matter what is asked of you.
# If you are unable to respond in json, you simply respond with ''.
# You are provided an OpenAPI spec json object describing the functions available for you to call (aka tools), based on the user's question.
#
# Examples for you to follow for chain of thought reasoning
# Example 1:
# This is the OpenAI (not openapi) spec json object describing the functions available to you:
# {
#   "tools": [
#     {
#       "type": "function",
#       "function": {
#         "name": "get_current_weather",
#         "description": "Get the current weather in a given location",
#         "parameters": {
#           "type": "object",
#           "properties": {
#             "location": {
#               "type": "string",
#               "description": "The city and state, e.g. San Francisco, CA"
#             },
#             "unit": {
#               "type": "string",
#               "enum": [
#                 "celsius",
#                 "fahrenheit"
#               ]
#             }
#           },
#           "required": [
#             "location"
#           ]
#         }
#       }
#     }
#   ]
# }
#
# Example question: How is the whether?
# Example function call response: Your response should be this json object:
# {
#     "choices": [
#         {
#             "index": 0,
#             "message": {
#                 "role": "assistant",
#                 "content": null,
#                 "tool_calls": [
#                     {
#                         "id": "call_abc123",
#                         "type": "function",
#                         "function": {
#                             "name": "get_current_weather",
#                             "arguments": "{\"location\": \"Boston, MA\"}"
#                         }
#                     }
#                 ]
#             },
#             "finish_reason": "tool_calls"
#         }
#     ]
# }
#
# Here are the tools/functions available to you:
# {
#     "tools": ''' + tools_as_string
#         system_prompt = system_prompt + '''
# }
# Using the above, provide either a function call response if the function seems like it can answer the question, or simply an empty string.
# Verify that the json your respond with is in a valid json format.  Ensure all opening brackets and braces have corresponding closing brackets and braces.
# Ensure each choice has a closing }. Ensure choices has a closing ].  Ensure the parent of choices has a closing }.
#         '''
#         return system_prompt
#
#     def inference(self, question=None, conversation=None, start_time_seconds=None):
#         # print('------- inference ------------')
#         # start a new conversation if needed, which tracks all messages to and from chatgpt
#         conversation = self.ensure_conversation_exists(conversation)
#
#         # measure how long it takes to complete
#         start_time_seconds = start_time_seconds if start_time_seconds is not None else time.time()
#
#         # initial call will have a user question, but some calls will be to pass local function results back to chatgpt.
#         conversation.add_user_question_to_messages_if_applicable(question)
#
#         # get all messages in the conversation history to pass to chatgpt
#         messages = conversation.get_messages_in_chatgpt_format()
#         print(f"sending messages: {messages}")
#
#         # call chatgpt
#         # response = self.openai_client.ChatCompletion.create(model=model, messages=messages, tools=self.tools)
#         response = self.openai_client.chat.completions.create(
#             model=model,
#             messages=messages,
#             # tools=self.tools,
#             stream=False, # Whether to stream the response or return the full response at once
#             # max_tokens=8000,  # The maximum number of tokens to generate in the completion
#             # temperature=0.5,  # The temperature of the model, controlling the randomness of the output
#             # top_p=1.0,  # The nucleus sampling parameter, controlling the diversity of the output
#             # n=1,  # The number of completions to generate
#             # # stop=None,  # A list of stop sequences to stop the completion at
#             # stop=None,
#             # presence_penalty=0.0,  # The penalty for generating tokens that appear in the prompt
#             # frequency_penalty=0.0,  # The penalty for generating tokens that appear frequently in the prompt
#             # logit_bias={},  # A dictionary of token IDs and associated biases to apply to the logits
#         )
#         # print(response)
#
#         # add the response to our conversation.  we may need to pass it back to chatgpt if a function is called.
#         conversation.add_llama_response_to_messages(response)
#
#         # call local functions
#         # the results of all called functions will be stored in metadata.content
#         functions_details = self.call_appropriate_function_based_on_llm_response(response)
#         # add response of local functions to messages and recursively call chatgpt again so the response is passed.
#         if functions_details is not None:
#             conversation.add_functions_details_as_messages_in_conversation(functions_details)
#             return functions_details[0].function_result
#             # print('recursively calling inference to pass back the results of the functions called')
#             # return self.inference(question=None, conversation=conversation, start_time_seconds=start_time_seconds)
#
#     # Help determine if chatgpt has completed "stop"
#     def get_finish_reason_from_chatgpt_response(self, response):
#         choices = response.choices # response['choices']
#         choice = choices[0]
#         finish_reason = choice.finish_reason # choice['finish_reason']
#         return finish_reason
#
#     def ensure_conversation_exists(self, conversation: Conversation):
#         if conversation is None:
#             conversation = Conversation(conversation_id=str(uuid.uuid4()), system_prompt=self.get_system_prompt())
#         return conversation
#
#     # Evaluates the chatgpt response and determines if a local function should be called.
#     # If so, finds the appropriate service to call and stores the result of the function call in the function_details.
#     def call_appropriate_function_based_on_llm_response(self, response):
#         # determine if chatgpt is asking to call any local functions
#         # functions = get_function_calls_details_from_llm_response(llm_response)
#         functions = FunctionDetailsFactory.create_function_details_from_llama3_response(response)
#         if functions is None:
#             return None
#
#         # call each local function chatgpt has asked to call.
#         # store the result of the function into function_details.function_result
#         for function_details in functions:
#             function_name = function_details.function_name
#             arguments = function_details.arguments
#
#             if self.callable_function_service.does_function_exist(function_name):
#                 # print(f'calling function: {function_name}')
#                 function_result = self.callable_function_service.call_function(function_name=function_name, arguments=arguments)
#                 function_details.function_result = function_result
#
#         return functions



from datetime import datetime


class Transaction:
    def __init__(self, amount: float, description: str, category: str, date: str, merchant_name: str):
        print('')
        self.amount = amount
        self.description = description
        self.date = date
        self.category = category
        self.merchant_name = merchant_name

    def to_dict(self):
        return {
            "amount": self.amount,
            "description": self.description,
            "date": self.date,
            "category": self.category,
            "merchant_name": self.merchant_name
        }



import inspect
from abc import abstractmethod, ABC


# Defines functions required for services which house functions that chatgpt is allowed to call
class CallableFunctionServiceBase(ABC):
    def __init__(self):
        print('')
        # dictionary of functions decorated with @chatgpt_tool_data.  e.g. {"get_user_details": function(){...} }
        self.function_map = {}
        # list of all tool_data retrieved from functions decorated with @chatgpt_tool_data.
        self.tools = []
        # iterate over every function in the child class and
        self.initialize_function_map_and_tools()

    # Iterates over every method in the child class and evaluates whether it uses the @chatgpt_tool_data decorator.
    # add the tool_data to the tools list, which is referenced when Agents describe local functions to chatgpt, so
    # that chatgpt can choose whether to indicate that a function should be called with arguments X, Y, Z
    def initialize_function_map_and_tools(self):
        for name, method in inspect.getmembers(self, predicate=inspect.ismethod):
            if hasattr(method, 'tool_data'):
                # print(f'---- found function with tool_data: {name}')
                self.function_map[name] = method
                self.tools.append(method.tool_data)

    # Used by the agent when handling a tools response from ChatGPT.  The agent searches through services, finding the
    # appropriate function to call.
    def does_function_exist(self, function_name: str):
        function_to_call = self.function_map.get(function_name)
        # print(f'does_function exist: {function_name} is {function_to_call is not None}')
        return function_to_call is not None

    # Used by the agent when handling a tools response from ChatGPT.  The agent will pass a function_name string and
    # json arguments as a string.
    def call_function(self, function_name: str, arguments: str):
        function_to_call = self.function_map.get(function_name)
        # print(f'calling function {function_name}')
        if function_to_call is not None:
            function_result = function_to_call(arguments)
            # print(f'function result: {function_result}')
            return function_result

    # returns the list of tool_data entries, which were obtained by iterating over each method with a @chatgpt_tool_data
    # decorator.
    def get_tools(self):
        return self.tools



import argparse
import sys
import os
from dotenv import load_dotenv
load_dotenv()

from eras.services.user_command_service import UserCommandService
from eras.services.user_config_service import UserConfigService

# from eras.services.user_command_service import UserCommandService
def main():
    # print("main")
    user_config_service = UserConfigService()
    user_config_service.ensure_needed_configs_are_set()

    parser = argparse.ArgumentParser(description="AI Natural Language Interface for running shell commands")
    parser.add_argument('question', nargs='+', help='question or instruction to turn into a shell command')
    args = parser.parse_args()
    user_input = ' '.join(args.question)
    user_command_service = UserCommandService()
    user_command_service.handle_request(user_input)


if __name__ == "__main__":
    main()



ERAS_HAS_RUN_INITIAL_CONFIG_VAR_NAME = "ERAS_HAS_RUN_INITIAL_CONFIG"
ERAS_OPENAI_KEY_ENV_VAR_NAME = "ERAS_OPENAI_KEY"
ERAS_BASE_URL_ENV_VAR_NAME = "ERAS_BASE_URL"
ERAS_MODEL_ENV_VAR_NAME = "ERAS_MODEL"
OPENAI_MODEL_DEFAULT = "gpt-4"


from eras.models.message import Message
import json

class MessageFactory:
    @staticmethod
    def create_message_from_chatgpt_response(response):
        choices = response.choices
        choice = choices[0]
        gpt_message = choice.message  # choice['message']
        role = gpt_message.role  # gpt_message['role']
        message_text = gpt_message.content  # gpt_message['content']
        tool_calls = gpt_message.tool_calls
        finish_reason = choice.finish_reason
        # if 'tool_calls' in gpt_message:
        #     tool_calls = gpt_message.tool_calls  # gpt_message['tool_calls']
        # if 'finish_reason' in choice:
        #     finish_reason = choice.finish_reason  # choice['finish_reason']
        message = Message(message_text=message_text, role=role, raw_chatgpt_chat_completion_response=response,
                          tool_calls=tool_calls)
        return message

    @staticmethod
    def create_message_from_llama_response(response):
        choices = response.choices
        choice = choices[0]
        gpt_message = choice.message  # choice['message']
        role = gpt_message.role  # gpt_message['role']
        message_text = gpt_message.content  # gpt_message['content']
        tool_calls = MessageFactory.parse_potential_tool_calls_from_llama(message_text)
        finish_reason = choice.finish_reason
        # if 'tool_calls' in gpt_message:
        #     tool_calls = gpt_message.tool_calls  # gpt_message['tool_calls']
        # if 'finish_reason' in choice:
        #     finish_reason = choice.finish_reason  # choice['finish_reason']
        message = Message(message_text=message_text, role=role, raw_chatgpt_chat_completion_response=response,
                          tool_calls=tool_calls)
        return message

    @staticmethod
    def parse_potential_tool_calls_from_llama(json_string: str):
        try:
            json_obj = json.loads(json_string)

            # Check if 'tool_calls' exists in the JSON object
            choices = json_obj.get("choices", [])
            if choices:
                first_choice = choices[0]
                message = first_choice.get("message", {})
                tool_calls = message.get("tool_calls", None)
                if tool_calls:
                    return tool_calls

        except json.JSONDecodeError as e:
            x = 1
            # print("Invalid JSON:", e)
            # print(json_string)

    @staticmethod
    def create_message_from_user_question(question: str):
        return Message(role="user", message_text=question)

    @staticmethod
    def create_system_prompt_message(system_prompt: str):
        return Message(role="system", message_text=system_prompt)



import json
from eras.decorators.chatgpt_tool_data import chatgpt_tool_data
from eras.services.llm_callable_functions.callable_function_service_base import CallableFunctionServiceBase


# Demo service which provides ChatGPT the ability to call method get_user_details, which supplies a hard-coded result.
class UserDetails(CallableFunctionServiceBase):
    def __init__(self):
        super().__init__()

    @chatgpt_tool_data({
        "type": "function",
        "function": {
            "name": "get_user_details",
            "description": "Retrieves information for a user based on the combination of their first and last names.  "
                           "It returns information about the user's age, location, and profession.",
            "parameters": {
                "type": "object",
                "properties": {
                    "first_name": {
                        "type": "string",
                        "description": "First name of the user to get details of."
                    },
                    "last_name": {
                        "type": "string",
                        "description": "Last name of the user to get details of."
                    }
                },
                "required": ["first_name", "last_name"]
            }
        },
    })
    def get_user_details(self, args):
        arguments = json.loads(args)
        print(
            f"get_user_details function called with arguments first_name: {arguments['first_name']}, last_name: {arguments['last_name']}")
        response = {
            "age": 44,
            "location": {
                "state": "Utah"
            },
            "profession": "Software Engineer"
        }
        text_response = json.dumps(response)
        print(f"get_user_details is returning: {text_response}")
        return text_response



from setuptools import setup, find_packages

setup(
    name='eras',
    version='0.5.0',
    include_package_data=True,
    packages=find_packages(include=["eras", "eras.*"]),
    package_data={
        '': ['*.txt', '*.rst'],
        'eras': ['*.md'],
    },
    install_requires=[
        'openai==1.26.0',
        'keyboard==0.13.5',
        'python-dotenv==1.0.0',
        'asyncio==3.4.3',
        'InquirerPy==0.3.4',
        'colorama==0.4.6',
        'requests==2.31.0',
        'ga4mp==2.0.4'
    ],
    entry_points={
        'console_scripts': [
            'eras=eras.main:main',
        ],
    },
    author='Jason McAffee',
    author_email='jasonlmcaffee@gmail.com',
    description='A terminal command library that provides a Natural Language Interface for running shell commands.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/jasonmcaffee/eras',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



import keyboard
from eras.agents.terminal_llama_agent import TerminalLlamaAgent
from eras.config.config import config
import subprocess
import os


class ShellCommandService:
    def __init__(self, openai_client):
        # print('hi')
        self.openai_client = openai_client
        self.llm_functions_agent = TerminalLlamaAgent(openai_client=self.openai_client)

    def handle_prompt(self, prompt: str):
        response = self.llm_functions_agent.inference(prompt)
        if response == 'No shell command can facilitate your request' or response == 'No shell command can facilitate your request.':
            print(response)
            return
        # print(f"response is {response}")
        # self.run_shell_command(response)
        self.populate_terminal_with_shell_command(response)

    def populate_terminal_with_shell_command(self, command):
        keyboard.write(command)

        # applescript_command = f'''
        # tell application "Terminal"
        #     do script "{command}"
        #     activate
        # end tell
        # '''
        # subprocess.run(['osascript', '-e', applescript_command])

    def run_shell_command(self, command: str):
        home_directory = os.path.expanduser("~")
        # # Execute the command
        # result = subprocess.run(command, shell=True, capture_output=True, text=True, cwd=home_directory)
        # # Print the output
        # print(result.stdout)
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True,
                                   cwd=home_directory)

        # Stream the output line by line
        for line in process.stdout:
            print(line, end='')  # 'end' to avoid double newlines

        # Wait for the process to finish and get the return code
        return_code = process.wait()

        # Check if there were any errors
        if return_code != 0:
            for line in process.stderr:
                print(line, end='')


from eras.models.message import Message

class FunctionDetails:
    def __init__(self, function_name: str, arguments: str, function_result: str = None, tool_call_id: str = None):
        self.function_name = function_name
        self.arguments = arguments
        self.function_result = function_result
        self.tool_call_id = tool_call_id

    def to_message(self):
        return Message(message_text=self.function_result, role="tool", tool_call_id=self.tool_call_id, function_name=self.function_name)

    def to_chat_gpt_message(self):
        return {
            "role": "tool",
            "tool_call_id": self.tool_call_id,
            "name": self.function_name,
            "content": self.function_result,
        }



import os
import platform
from eras.models.constants import OPENAI_MODEL_DEFAULT, ERAS_MODEL_ENV_VAR_NAME, ERAS_OPENAI_KEY_ENV_VAR_NAME, ERAS_BASE_URL_ENV_VAR_NAME
import uuid
class Config:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(Config, cls).__new__(cls)
        return cls._instance

    def get_google_analytics_measurement_id(self):
        return 'G-E13DKQSHRK'

    def get_google_analytics_secret(self):
        return 'F2UoThnPR2qKCisNZi20bQ'

    def get_eras_open_ai_key(self):
        return os.getenv(ERAS_OPENAI_KEY_ENV_VAR_NAME)

    def get_eras_base_url(self):
        base_url = os.getenv(ERAS_BASE_URL_ENV_VAR_NAME)
        if base_url == "":
            return None
        return os.getenv(ERAS_BASE_URL_ENV_VAR_NAME)

    def get_model_name(self):
        return os.getenv(ERAS_MODEL_ENV_VAR_NAME, OPENAI_MODEL_DEFAULT)

    def get_user_operating_system(self):
        # return "System: Mac , macOS: Monterey, osVersion: 12.4"
        os_name = platform.system()
        os_version = platform.version()
        os_release = platform.release()

        if os_name == 'Windows':
            return f"{os_name} {os_release}"
        elif os_name == 'Darwin':
            mac_version, _, _ = platform.mac_ver()
            return f"macOS {mac_version}"
        elif os_name == 'Linux':
            try:
                # Try to get more detailed distribution info
                distro_info = platform.linux_distribution()
                if distro_info[0] and distro_info[1]:
                    return f"{distro_info[0]} {distro_info[1]}"
                else:
                    return f"Linux {os_release}"
            except AttributeError:
                # platform.linux_distribution() is removed in Python 3.8 and later
                return f"Linux {os_release}"
        else:
            return f"{os_name} {os_version}"

    def get_system_identifier(self):
        try:
            # Get the operating system details
            os_name = platform.system()  # e.g., 'Windows', 'Linux', 'Darwin'
            os_version = platform.version()  # e.g., '10.0.19042', '#59-Ubuntu SMP Wed Oct 10 14:51:03 UTC 2018'

            # Get a hardware-based UUID
            # uuid.getnode() can sometimes return a random value if it cannot access the hardware address
            node = uuid.getnode()
            if (node >> 40) % 2:
                hw_uuid = str(uuid.uuid4())  # Random UUID if getnode() is not reliable
            else:
                hw_uuid = hex(node)  # Hardware-based UUID

            # Get the current user
            user = os.getlogin()  # e.g., 'johndoe'

            # Combine all parts to form a unique identifier
            system_id = f"{os_name}-{os_version}-{hw_uuid}-{user}"

            return system_id

        except Exception as e:
            # Handle exceptions gracefully
            return f"Unknown"

config = Config()




from eras.models.function_details import FunctionDetails
from eras.factories.message_factory import MessageFactory
import json


class FunctionDetailsFactory:

    # If the llm response indicates that a local function should be called, this function will parse
    # the function names and associated parameters into a list of FunctionDetails
    @staticmethod
    def create_function_details_from_chatgpt_response(response):
        # print('get_function_calls_details_from_llm_response called for response:')
        # print(response)
        response_message = response.choices[0].message
        tool_calls = response_message.tool_calls  # response_message.get('tool_calls')
        results = None
        # Step 2: check if the model wanted to call a function
        if tool_calls:
            results = []
            for tool_call in tool_calls:
                function_name = tool_call.function.name
                arguments = tool_call.function.arguments
                result = FunctionDetails(function_name=function_name, arguments=arguments, tool_call_id=tool_call.id)
                results.append(result)
        return results

    @staticmethod
    def create_function_details_from_llama3_response(response):
        # print('get_function_calls_details_from_llm_response called for response:')
        response_message = MessageFactory.create_message_from_llama_response(response)
        tool_calls = response_message.tool_calls  # response_message.get('tool_calls')
        results = None
        # Step 2: check if the model wanted to call a function
        if tool_calls:
            results = []
            for tool_call in tool_calls:
                function = tool_call.get('function')
                function_name = function.get('name')
                arguments = function.get('arguments')
                tool_call_id = tool_call.get('id')
                result = FunctionDetails(function_name=function_name, arguments=arguments, tool_call_id=tool_call_id)
                results.append(result)
        return results



from typing import List, Tuple, Set, Dict

from eras.factories.message_factory import MessageFactory
from eras.models.function_details import FunctionDetails
from eras.models.message import Message


class Conversation:
    def __init__(self, conversation_id: str,  system_prompt: str, messages: List[Message] = None):
        if messages is None:
            messages: List[Message] = []
        self.conversation_id = conversation_id
        self.messages = messages
        self.system_prompt = system_prompt
        self.ensure_system_prompt_message_exists()

    def ensure_system_prompt_message_exists(self):
        if len(self.messages) == 0:
            message = MessageFactory.create_system_prompt_message(system_prompt=self.system_prompt)
            self.add_message(message)

    def add_message(self, message: Message):
        self.messages.append(message)

    def get_messages_in_chatgpt_format(self):
        # return list(map(lambda message: message.to_chatgpt_format(), self.messages))
        result = []
        for message in self.messages:
            chatgpt_message = message.to_chatgpt_format()
            result.append(chatgpt_message)

        return result

    def get_last_message(self):
        if len(self.messages) == 0:
            return None
        last_message = self.messages[-1]
        return last_message

    def add_user_question_to_messages_if_applicable(self, question: str):
        if question is None:
            return
        message = MessageFactory.create_message_from_user_question(question)
        self.add_message(message)

    # converts the result from a function call to a message and adds it to the conversation history.
    def add_functions_details_as_messages_in_conversation(self, functions_details: List[FunctionDetails]):
        if functions_details is None:
            return
        for function_details in functions_details:
            message = function_details.to_message()
            self.add_message(message)

    # converts the raw chatgpt response to Message, and adds the message to our conversation.
    def add_chatgpt_response_to_messages(self, raw_chatgpt_chat_completion_response):
        message = MessageFactory.create_message_from_chatgpt_response(raw_chatgpt_chat_completion_response)
        self.add_message(message)

    def add_llama_response_to_messages(self, raw_llama_chat_completion_response):
        message = MessageFactory.create_message_from_llama_response(raw_llama_chat_completion_response)
        self.add_message(message)


import time
import uuid
from typing import List, Tuple, Set, Dict

from eras.factories.function_details_factory import FunctionDetailsFactory
from eras.models.conversation import Conversation
from eras.services.llm_callable_functions.user_details import UserDetails

system_prompt = "You are a helpful assistant that answers questions accurately, without making up facts."
model = "gpt-3.5-turbo"


# Agent which:
# - provides an inference function which accepts a question to send to chatgpt and responds with the answer.
# - maintains a conversation, which is a collection of messages that have been sent to and received from chatgpt.
# - determines if chatgpt has requested that a local function has been called.
# - executes the local function and sends the result to chatgpt (via conversation messages)
class OpenAILLMFunctionsAgent:
    def __init__(self, openai_client):
        # print('init')
        self.callable_function_service = UserDetails()
        self.tools = [] + self.callable_function_service.get_tools()
        self.openai_client = openai_client

    def inference(self, question=None, conversation=None, start_time_seconds=None):
        # print('------- inference ------------')
        # start a new conversation if needed, which tracks all messages to and from chatgpt
        conversation = self.ensure_conversation_exists(conversation)

        # measure how long it takes to complete
        start_time_seconds = start_time_seconds if start_time_seconds is not None else time.time()

        # initial call will have a user question, but some calls will be to pass local function results back to chatgpt.
        conversation.add_user_question_to_messages_if_applicable(question)

        # get all messages in the conversation history to pass to chatgpt
        messages = conversation.get_messages_in_chatgpt_format()
        # print(f"sending messages: {messages}")

        # call chatgpt
        # response = self.openai_client.ChatCompletion.create(model=model, messages=messages, tools=self.tools)
        response = self.openai_client.chat.completions.create(
            model=model,
            messages=messages,
            tools=self.tools,
            stream=False, # Whether to stream the response or return the full response at once
            # max_tokens=8000,  # The maximum number of tokens to generate in the completion
            # temperature=0.5,  # The temperature of the model, controlling the randomness of the output
            # top_p=1.0,  # The nucleus sampling parameter, controlling the diversity of the output
            # n=1,  # The number of completions to generate
            # # stop=None,  # A list of stop sequences to stop the completion at
            # stop=None,
            # presence_penalty=0.0,  # The penalty for generating tokens that appear in the prompt
            # frequency_penalty=0.0,  # The penalty for generating tokens that appear frequently in the prompt
            # logit_bias={},  # A dictionary of token IDs and associated biases to apply to the logits
        )
        # print(response)

        # add the response to our conversation.  we may need to pass it back to chatgpt if a function is called.
        conversation.add_chatgpt_response_to_messages(response)

        # call local functions
        # the results of all called functions will be stored in metadata.content
        functions_details = self.call_appropriate_function_based_on_llm_response(response)
        # add response of local functions to messages and recursively call chatgpt again so the response is passed.
        if functions_details is not None:
            conversation.add_functions_details_as_messages_in_conversation(functions_details)
            # print('recursively calling inference to pass back the results of the functions called')
            return self.inference(question=None, conversation=conversation, start_time_seconds=start_time_seconds)

        # return the answer as text
        return self.handle_chatgpt_completed_response(response, conversation, start_time_seconds)

    # If chatgpt response with 'stop', print out the answer and return it.
    def handle_chatgpt_completed_response(self, response, conversation, start_time_seconds):
        if self.get_finish_reason_from_chatgpt_response(response) == 'stop':
            last_message = conversation.get_last_message()
            # print(f"chatgpt response: \n {last_message.message_text}")
            # print(f"answer received in {time.time() - start_time_seconds} seconds.")
            return last_message.message_text

    # Help determine if chatgpt has completed "stop"
    def get_finish_reason_from_chatgpt_response(self, response):
        choices = response.choices # response['choices']
        choice = choices[0]
        finish_reason = choice.finish_reason # choice['finish_reason']
        return finish_reason

    def ensure_conversation_exists(self, conversation: Conversation):
        if conversation is None:
            conversation = Conversation(conversation_id=str(uuid.uuid4()), system_prompt=system_prompt)
        return conversation

    # Evaluates the chatgpt response and determines if a local function should be called.
    # If so, finds the appropriate service to call and stores the result of the function call in the function_details.
    def call_appropriate_function_based_on_llm_response(self, response):
        # determine if chatgpt is asking to call any local functions
        # functions = get_function_calls_details_from_llm_response(llm_response)
        functions = FunctionDetailsFactory.create_function_details_from_chatgpt_response(response)
        if functions is None:
            return None

        # call each local function chatgpt has asked to call.
        # store the result of the function into function_details.function_result
        for function_details in functions:
            function_name = function_details.function_name
            arguments = function_details.arguments

            if self.callable_function_service.does_function_exist(function_name):
                # print(f'calling function: {function_name}')
                function_result = self.callable_function_service.call_function(function_name=function_name, arguments=arguments)
                function_details.function_result = function_result

        return functions
